#!/bin/sh

# Poisson Media Recovery
# 
# The script is in public domain. It was originally written by
# Heikki Orsila <heikki.orsila@iki.fi> at 2005-06-20.

# This script tries to obtain a broken file from cdrom devices by repeatedly
# reading blocks from the media.
# It sequentially mounts and umounts the device, and on each mount it tries
# to read as many new 512 byte sectors as possible.
# 
# I was able to use this tool to save some files from old CDs.

version="0.03"

mountpath="$1"
filepath="$2"
resultfile="$3"


function error {
    echo "fatal error:" "$@"
    exit -1
}


if test -z "$1" ; then
    cat <<EOF
poisson media recovery $version by Heikki Orsila <heikki.orsila@iki.fi>
Web site: http://zakalwe.fi/~shd/foss/poisson/
Git repository: git://zakalwe.fi/scripts.git

USAGE:

  poisson_media_recovery.sh DEVICEMOUNT FULL_FILE_PATH RESULT_FILE

DEVICEMOUNT (e.g. /mnt/cdrom) must be a valid mount path (see /etc/fstab).
FULL_FILE_PATH (e.g. /mnt/cdrom/myfile) is an absolute path into the file
to be recovered. RESULT_FILE (e.g. /tmp/myfile) is the path to the file
that needed to be recovered.

Requirements:

1. /mnt/cdrom must not be mounted initially
2. There must be enough free space at /tmp to hold largest file on the cdrom

EXAMPLE:

poisson_media_recovery.sh /mnt/cdrom /mnt/cdrom/mybrokenfile /tmp/savedfile

EOF
    exit 1
fi

if test -e "$resultfile" ; then error "$resultfile exists" ; fi

tmpfile="$(tempfile)" || error "Can not create temprorary file"

echo "Using $tmpfile as a temporary file..."

size="0"
pass="1"

while true ; do

    echo Pass $pass

    mount "$mountpath"
    if test "$?" != "0" ; then
	rm -f "$tmpfile"
	error "Can not mount $mountpath. Maybe it was already mounted or is not a cdrom mount point at all?"
    fi

    # Get total file size
    fullsize="$(stat -c "%s" "$filepath")"

    if test -z "$fullsize" ; then
	rm -f "$tmpfile"
	umount "$mountpath"
	error "Can not stat $filepath"
    fi

    echo "Reading at offset $size"

    # Read as many 512 byte blocks as possible
    dd if="$filepath" bs=512 skip=$((size / 512)) > "$tmpfile" 2>/dev/null
    retval="$?"

    umount "$mountpath"

    if test "$retval" = "0" ; then
	# Great, we were able to read everything -> job done
	cat "$tmpfile" >> "$resultfile"
	break
    else
	echo "Read error detected"

	# Get file size
	newsize="$(stat -c "%s" "$tmpfile")"

	if test -z "$newsize" ; then
	    rm -f "$tmpfile"
	    error "Can not stat $tmpfile (very odd, please report this bug)"
	fi

	addition=$((newsize / 512))

	if test "$addition" -gt "0" ; then
	    # Got at least 512 bytes
	    echo "Addition of $addition * 512 bytes"
	    size=$((size + addition * 512))
	    dd if="$tmpfile" bs=512 count="$addition" >> "$resultfile" 2>/dev/null
	    echo "The new recovered size is $size bytes"
	else
	    # Got less than 512 bytes.
	    # Test if this is the last batch of bytes

	    remainder=$((fullsize - size))

	    if test "$newsize" -eq "$remainder" ; then
		cat "$tmpfile" >> "$resultfile"
		break
	    fi

	    echo "No new data blocks. Retrying.."
	fi
    fi

    pass=$((pass + 1))
done

rm -f "$tmpfile"
